@@ -2,14 +2,14 @@
 
 import os
 import socket
-from stat import S_IMODE
+from stat import S_IMODE, S_ISDIR, S_ISREG
 import tempfile
 import paramiko
 from paramiko import SSHException   # make available
 from paramiko import AuthenticationException   # make available
 from paramiko import AgentKey
 
-__version__ = "0.2.5"
+__version__ = "0.2.6"
 
 
 def st_mode_to_int(val):
@@ -146,7 +146,8 @@ class Connection(object):
             self._sftp = paramiko.SFTPClient.from_transport(self._transport)
             self._sftp_live = True
 
-    def get(self, remotepath, localpath=None, callback=None):
+    def get(self, remotepath, localpath=None, callback=None,
+            preserve_mtime=False):
         """Copies a file between the remote host and the local host.
 
         :param str remotepath: the remote path and filename, source
@@ -156,6 +157,10 @@ class Connection(object):
         :param callable callback:
             optional callback function (form: ``func(int, int)``) that accepts
             the bytes transferred so far and the total bytes to be transferred.
+        :param bool preserve_mtime:
+            *Default: False* - make the modification time(st_mtime) on the
+            local file match the time on the remote. (st_atime can differ
+            because stat'ing the localfile can/does update it's st_atime)
 
         :returns: nothing
 
@@ -164,8 +169,14 @@ class Connection(object):
         """
         if not localpath:
             localpath = os.path.split(remotepath)[1]
+
         self._sftp_connect()
+        if preserve_mtime:
+            sftpattrs = self._sftp.stat(remotepath)
+
         self._sftp.get(remotepath, localpath, callback=callback)
+        if preserve_mtime:
+            os.utime(localpath, (sftpattrs.st_atime, sftpattrs.st_mtime))
 
     def getfo(self, remotepath, flo, callback=None):
         """Copy a remote file (remotepath) to a file-like object, flo.
@@ -184,7 +195,8 @@ class Connection(object):
         self._sftp_connect()
         return self._sftp.getfo(remotepath, flo, callback=callback)
 
-    def put(self, localpath, remotepath=None, callback=None, confirm=True):
+    def put(self, localpath, remotepath=None, callback=None, confirm=True,
+            preserve_mtime=False):
         """Copies a file between the local host and the remote host.
 
         :param str localpath: the local path and filename
@@ -196,6 +208,10 @@ class Connection(object):
         :param bool confirm:
             whether to do a stat() on the file afterwards to confirm the file
             size
+        :param bool preserve_mtime:
+            *Default: False* - make the modification time(st_mtime) on the
+            remote file match the time on the local. (st_atime can differ
+            because stat'ing the localfile can/does update it's st_atime)
 
         :returns:
             SFTPAttributes object containing attributes about the given file
@@ -206,8 +222,18 @@ class Connection(object):
         if not remotepath:
             remotepath = os.path.split(localpath)[1]
         self._sftp_connect()
-        return self._sftp.put(localpath, remotepath, callback=callback,
-                              confirm=confirm)
+
+        if preserve_mtime:
+            local_stat = os.stat(localpath)
+            times = (local_stat.st_atime, local_stat.st_mtime)
+
+        sftpattrs = self._sftp.put(localpath, remotepath, callback=callback,
+                                   confirm=confirm)
+        if preserve_mtime:
+            self._sftp.utime(remotepath, times)
+            sftpattrs = self._sftp.stat(remotepath)
+
+        return sftpattrs
 
     def putfo(self, flo, remotepath=None, file_size=0, callback=None,
               confirm=True):
@@ -268,6 +294,48 @@ class Connection(object):
         self._sftp_connect()
         self._sftp.chdir(remotepath)
 
+    def chmod(self, remotepath, mode=777):
+        """set the mode of a remotepath to mode, where mode is an integer
+        representation of the octal mode to use.
+
+        :param str remotepath: the remote path/file to modify
+        :param int mode:
+            int representation of octal mode for directory, default 777
+
+        :returns: None
+
+        :raises: IOError if the file doesn't exist
+
+        """
+        self._sftp_connect()
+        self._sftp.chmod(remotepath, mode=int(str(mode), 8))
+
+    def chown(self, remotepath, uid=None, gid=None):
+        """ set uid and/or gid on a remotepath, you may specify either or both.
+        Unless you have **permission** to do this on the remote server, you will
+        raise an IOError: 13 - permission denied
+
+        :param str remotepath: the remote path/file to modify
+        :param int uid: the user id to set on the remotepath
+        :param int gid: the group id to set on the remotepath
+
+        :returns None:
+
+        :raises IOError: if you don't have permission or the file doesn't exist
+
+        """
+        self._sftp_connect()
+        if uid is None or gid is None:
+            if uid is None and gid is None:  # short circuit if no change
+                return
+            rstat = self._sftp.stat(remotepath)
+            if uid is None:
+                uid = rstat.st_uid
+            if gid is None:
+                gid = rstat.st_gid
+
+        self._sftp.chown(remotepath, uid=uid, gid=gid)
+
     def getcwd(self):
         """return the current working directory on the remote
 
@@ -303,6 +371,68 @@ class Connection(object):
         self._sftp_connect()
         self._sftp.mkdir(remotepath, mode=int(str(mode), 8))
 
+    def isdir(self, remotepath):
+        """return true if remotepath is a directory
+
+        :param str remotepath: the path to test
+
+        :returns bool:
+
+        """
+        self._sftp_connect()
+        try:
+            result = S_ISDIR(self._sftp.stat(remotepath).st_mode)
+        except IOError:     # no such file
+            result = False
+        return result
+
+    def isfile(self, remotepath):
+        """return true if remotepath is a file
+
+        :param str remotepath: the path to test
+
+        :returns bool:
+
+        """
+        self._sftp_connect()
+        try:
+            result = S_ISREG(self._sftp.stat(remotepath).st_mode)
+        except IOError:     # no such file
+            result = False
+        return result
+
+    def makedirs(self, remotedir, mode=777):
+        """create all directories in remotedir as needed, setting their mode
+        to mode, if created.
+
+        If remotedir already exists, silently complete. If a regular file is
+        in the way, raise an exception.
+
+        :param str remotedir: the direcotry structure to create
+        :param int mode:
+            int representation of octal mode for directory, default 777
+
+        :returns: None
+
+        :raises: OSError
+
+        """
+        self._sftp_connect()
+        if self.isdir(remotedir):
+            pass
+
+        elif self.isfile(remotedir):
+            raise OSError("a file with the same name as the remotedir, " \
+                          "'%s', already exists." % remotedir)
+        else:
+
+            head, tail = os.path.split(remotedir)
+            if head and not self.isdir(head):
+                self.makedirs(head, mode)
+
+            if tail:
+                self.mkdir(remotedir, mode=mode)
+
     def remove(self, remotefile):
         """remove the file @ remotefile, remotefile may include a path, if no
         path, then cwd is used.  This method only works on files
@@ -395,6 +525,67 @@ class Connection(object):
         self._sftp_connect()
         return self._sftp.open(remote_file, mode=mode, bufsize=bufsize)
 
+    def exists(self, remotepath):
+        """Test whether a remotepath exists.
+
+        :param str remotepath: the remote path to verify
+
+        :returns bool: True if remotepath exists, else False
+
+        """
+        self._sftp_connect()
+        try:
+            self._sftp.stat(remotepath)
+        except IOError:
+            return False
+        return True
+
+    def lexists(self, remotepath):
+        """Test whether a remotepath exists.  Returns True for broken symbolic
+        links
+
+        :param str remotepath: the remote path to verify
+
+        :returns bool: True if lexists, else False
+
+        """
+        self._sftp_connect()
+        try:
+            self._sftp.lstat(remotepath)
+        except IOError:
+            return False
+        return True
+
+    def symlink(self, remote_src, remote_dest):
+        '''create a symlink for a remote file on the server
+
+        :param str remote_src: path of original file
+        :param str remote_dest: path of the created symlink
+
+        :returns: None
+
+        :raises:
+            any underlying error, IOError if something already exists at
+            remote_dest
+
+        '''
+        self._sftp_connect()
+        self._sftp.symlink(remote_src, remote_dest)
+
+    @property
+    def sftp_client(self):
+        """give access to the underlying, connected paramiko SFTPClient object
+
+        see http://paramiko-docs.readthedocs.org/en/latest/api/sftp.html?highlight=sftpclient
+
+        :params: None
+
+        :returns: the active SFTPClient object
+
+        """
+        self._sftp_connect()
+        return self._sftp
+
     @property
     def active_ciphers(self):
         """Get tuple of currently used local and remote ciphers.
@@ -410,9 +601,10 @@ class Connection(object):
         """return the available security options recognized by paramiko.
 
         :returns SecurityOptions:
-            a simple object security preferences of an ssh transport. These
-            are tuples of acceptable ciphers, digests, key types, and key
-            exchange algorithms, listed in order of preference.
+            a simple object security preferences of an
+            ssh transport. These are tuples of acceptable ciphers, digests,
+            key types, and key exchange algorithms, listed in order of
+            preference.
 
         """
 
@@ -437,3 +629,47 @@ class Connection(object):
     def __exit__(self, etype, value, traceback):
         self.close()
 
+
+def path_advance(thepath, sep=os.sep):
+    '''generator to iterate over a file path forwards
+
+    :param str thepath: the path to navigate forwards
+    :param str sep: the path separator to use, defaults to ``os.sep``
+
+    :returns generator: of strings
+
+    '''
+    # handle a direct path
+    pre = ''
+    if thepath[0] == sep:
+        pre = sep
+    curpath = ''
+    parts = thepath.split(sep)
+    if pre:
+        if parts[0]:
+            parts[0] = pre + parts[0]
+        else:
+            parts[1] = pre + parts[1]
+    for part in parts:
+        curpath = os.path.join(curpath, part)
+        if curpath:
+            yield curpath
+
+
+def path_retreat(thepath, sep=os.sep):
+    '''generator to iterate over a file path in reverse
+
+    :param str thepath: the path to retreat over
+    :param str sep: the path separator to use, default to ``os.sep``
+
+    :returns generator: of strings
+
+    '''
+    pre = ''
+    if thepath[0] == sep:
+        pre = sep
+    parts = thepath.split(sep)
+    while parts:
+        if os.path.join(*parts):
+            yield '%s%s' % (pre, os.path.join(*parts))
+        parts = parts[:-1]
